package com.example.application.views.abstractnavigation;

import com.example.application.data.entity.Matchday;
import com.example.application.data.entity.Season;
import com.example.application.data.service.MatchdayService;
import com.example.application.data.service.SeasonService;
import com.vaadin.flow.component.AbstractField;
import com.vaadin.flow.component.Component;
import com.vaadin.flow.component.HasValue;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.html.Div;
import com.vaadin.flow.component.html.Label;
import com.vaadin.flow.component.orderedlayout.FlexComponent;
import com.vaadin.flow.component.orderedlayout.HorizontalLayout;
import com.vaadin.flow.component.orderedlayout.VerticalLayout;
import com.vaadin.flow.component.select.Select;
import com.vaadin.flow.router.BeforeEvent;
import com.vaadin.flow.router.HasUrlParameter;
import com.vaadin.flow.router.WildcardParameter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.util.ArrayList;
import java.util.List;

public abstract class SeasonAndMatchdayNavigationView extends Div implements HasUrlParameter<String> {
    // TODO: show dropdown menus also for invalid URLs (with content that fits the situation)

    protected final SeasonService seasonService;
    protected final MatchdayService matchdayService;

    protected String seasonParam;
    protected String matchdayParam;

    private final Label invalidUrlLabel = new Label();

    protected final VerticalLayout outer = new VerticalLayout();
    protected final HorizontalLayout selectionLayout = new HorizontalLayout();

    protected final List<Season> seasonList = new ArrayList<>();
    protected final Select<Season> seasonSelect = new Select<>();

    protected final List<Matchday> matchdayList = new ArrayList<>();
    protected final Select<Matchday> matchdaySelect = new Select<>();
    protected Component contentLayout;

    public SeasonAndMatchdayNavigationView(@Autowired SeasonService seasonService, @Autowired MatchdayService matchdayService) {
        this.seasonService = seasonService;
        this.matchdayService = matchdayService;

        addClassName("results-view");

        configureOuterLayout();
        configureSelectionLayout();
        configureContentLayout();
    }

    protected abstract String route();

    protected abstract void configureContentLayout();

    protected abstract void configureContent();

    protected abstract boolean showOnlyMatchdaysWithActivity();

    private void configureOuterLayout() {
        add(outer);
        outer.setAlignItems(FlexComponent.Alignment.CENTER);
    }

    private void configureSelectionLayout() {
        selectionLayout.setWidthFull();
        selectionLayout.setAlignItems(FlexComponent.Alignment.CENTER);
        selectionLayout.setJustifyContentMode(FlexComponent.JustifyContentMode.END);
        selectionLayout.add(new Label("Season:"), seasonSelect, new Label("Matchday:"), matchdaySelect);

        seasonSelect.addValueChangeListener(seasonSelectValueChangeListener());
        matchdaySelect.addValueChangeListener(matchdaySelectValueChangeListener());

        // provide data
        fillSeasonSelectWithData();
    }

    private HasValue.ValueChangeListener<? super AbstractField.ComponentValueChangeEvent<Select<Matchday>, Matchday>> matchdaySelectValueChangeListener() {
        return matchdayChangeEvent -> getUI().ifPresent(ui -> {
            Season season = seasonSelect.getValue();
            Matchday matchday = matchdayChangeEvent.getValue();
            if (season != null && matchday != null) {
                String seasonParam = season.toString();
                String matchdayParam = matchday.toString();
                navigate(ui, seasonParam, matchdayParam);
            }
        });
    }

    private HasValue.ValueChangeListener<? super AbstractField.ComponentValueChangeEvent<Select<Season>, Season>> seasonSelectValueChangeListener() {
        return seasonChangeEvent -> getUI().ifPresent(ui -> {
            Season newSeason = seasonChangeEvent.getValue();
            if (newSeason != null) {
                String seasonParam = newSeason.toString();
                String matchdayParam = null;

                Matchday matchdayInNewSeason = null;
                Matchday matchdayInOldSeason = matchdaySelect.getValue();
                if (matchdayInOldSeason != null) {
                    matchdayParam = matchdayInOldSeason.toString();
                    matchdayInNewSeason = getMatchdayFromParam(matchdayParam, newSeason);
                }
                matchdayParam = matchdayInNewSeason == null ? "1" : matchdayParam;

                navigate(ui, seasonParam, matchdayParam);
            }
        });
    }

    private void fillSeasonSelectWithData() {
        seasonList.clear();
        seasonList.addAll(seasonService.getAllSeasonsSorted());
        seasonSelect.setItems(seasonList);
    }

    private void fillMatchdaySelectWithData() {
        matchdayList.clear();
        List<Matchday> matchdaysToAdd = showOnlyMatchdaysWithActivity() ?
                matchdayService.getMatchdaysWithActivityForSeasonSorted(seasonSelect.getValue()) : matchdayService.getMatchdaysForSeasonSorted(seasonSelect.getValue());
        matchdayList.addAll(matchdaysToAdd);
        matchdaySelect.setItems(matchdayList);
    }

    protected boolean isMatchDayParamValid(@NonNull String matchdayParam) {
        return matchdayList.stream().anyMatch(matchday -> matchdayParam.equals(matchday.toString()));
    }

    protected void navigate(UI ui, String seasonParam, String matchdayParam) {
        ui.navigate(String.format("%s/%s/%s/", route(), seasonParam, matchdayParam));
    }

    @Override
    public void setParameter(BeforeEvent beforeEvent, @WildcardParameter String param) {
        outer.removeAll();

        // TODO: handle corner case where first matchday has no active games yet: still show first matchday
        if (param.equals("")) {
            Season latestSeason = seasonService.getLatestSeason();
            if (latestSeason != null) {
                Matchday latestMatchday = matchdayService.getLastMatchdayWithActivityForSeason(latestSeason);
                if (latestMatchday != null) {
                    String seasonParam = latestSeason.toString();
                    String matchdayParam = latestMatchday.toString();
                    navigate(UI.getCurrent(), seasonParam, matchdayParam);
                    setParameter(beforeEvent, String.format("%s/%s/", seasonParam, matchdayParam)); // TODO: check why this line seems to be necessary
                }
            }
            return;
        }

        if (!param.matches("^[0-9]*/[0-9]*/?$")) {
            invalidUrlLabel.setText("Invalid URL! Please provide params in the form season/matchday/");
            outer.add(invalidUrlLabel);
            return;
        }

        String[] params = param.split("/");
        seasonParam = params[0];
        matchdayParam = params[1];

        Season season = getSeasonFromParam(seasonParam);
        if (season == null) {
            invalidUrlLabel.setText(String.format("Invalid URL! Season \"%s\" does not exist in the database!", seasonParam));
            outer.add(invalidUrlLabel);
            return;
        }

        seasonSelect.setValue(season);
        fillMatchdaySelectWithData();

        Matchday matchday = getMatchdayFromParam(matchdayParam);
        if (matchday == null) {
            String messageExtra = showOnlyMatchdaysWithActivity() ? " or has no games played yet" : "";
            invalidUrlLabel.setText(String.format("Invalid URL! Matchday \"%s\" in Season \"%s\" does not exist in the database%s!", matchdayParam, seasonParam, messageExtra));
            outer.add(invalidUrlLabel);
            return;
        }

        matchdaySelect.setValue(matchday);

        outer.add(selectionLayout);
        outer.add(contentLayout);

        configureContent();
    }

    @Nullable
    private Season getSeasonFromParam(@NonNull String seasonParam) {
        for (Season season : seasonList) {
            if (seasonParam.equals(season.toString())) {
                return season;
            }
        }
        return null;
    }

    @Nullable
    private Matchday getMatchdayFromParam(@NonNull String matchdayParam) {
        return getMatchdayFromParam(matchdayParam, null);
    }

    @Nullable
    private Matchday getMatchdayFromParam(@NonNull String matchdayParam, @Nullable Season season) {
        List<Matchday> matchdayList = season == null ? this.matchdayList : matchdayService.getMatchdaysForSeasonSorted(season);
        for (Matchday matchday : matchdayList) {
            if (matchdayParam.equals(matchday.toString())) {
                return matchday;
            }
        }
        return null;
    }
}
